# -*- coding: utf-8 -*-
'''base stuf'''

from __future__ import absolute_import
from collections import Mapping, Sequence

from stuf.util import class_name, lazy, recursive_repr


class basestuf(object):

    '''stuf basestuf'''

    _mapping = dict
    _reserved = ['allowed', '_wrapped', '_mapping']

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        self.update(*args, **kw)

    def __iter__(self):
        cls = self.__class__
        for k, v in self.iteritems():
            # nested stuf of some sort
            if isinstance(v, cls):
                yield (k, dict(i for i in v))
            # normal key, value pair
            else:
                yield (k, v)

    def __getattr__(self, k):
        try:
            return self.__getitem__(k)
        except KeyError:
            return object.__getattribute__(self, k)

    def __setattr__(self, k, v):
        # handle normal object attributes
        if k == '_classkeys' or k in self._classkeys:
            object.__setattr__(self, k, v)
        # handle special attributes
        else:
            try:
                self.__setitem__(k, v)
            except:
                raise AttributeError(k)

    def __delattr__(self, k):
        # allow deletion of key-value pairs only
        if not k == '_classkeys' or k in self._classkeys:
            try:
                self.__delitem__(k)
            except KeyError:
                raise AttributeError(k)

    def __getstate__(self):
        return self._mapping(self)

    @recursive_repr
    def __repr__(self):
        if not self:
            return '%s()' % class_name(self)
        return '%s(%r)' % (class_name(self), self.items())

    def __setstate__(self, state):
        return self._build(state)

    @lazy
    def _classkeys(self):
        # protected keywords
        return frozenset(
            vars(self).keys() + vars(self.__class__).keys() + self._reserved
        )

    @classmethod
    def _build(cls, iterable):
        kind = cls._mapping
        # add class to handle potential nested objects of the same class
        kw = kind()
        if isinstance(iterable, Mapping):
            kw.update(kind(i for i in iterable.items()))
        elif isinstance(iterable, Sequence):
            # extract appropriate key-values from sequence
            for arg in iterable:
                try:
                    kw.update(arg)
                except (ValueError, TypeError):
                    pass
        return kw

    @classmethod
    def _new(cls, iterable):
        return cls(cls._build(iterable))

    def _populate(self, iterable):
        new = self._new
        for k, v in iterable.iteritems():
            if isinstance(v, (Sequence, Mapping)):
                # see if stuf can be converted to nested stuf
                trial = new(v)
                if len(trial) > 0:
                    self.__setitem__(k, trial)
                else:
                    self.__setitem__(k, v)
            else:
                self.__setitem__(k, v)

    def _prepare(self, *args, **kw):
        kw.update(self._build(args))
        return kw

    def copy(self):
        return self._build(self._mapping(i for i in self))

    def update(self, *args, **kw):
        self._populate(self._prepare(*args, **kw))


class wrapstuf(basestuf):

    '''wraps mappings for stuf compliance'''

    def __getitem__(self, key):
        return self._wrapped[key]

    def __iter__(self):
        return self._wrapped.__iter__()

    def __len__(self):
        return self._wrapped.__len__()


class stuf(basestuf, dict):

    '''dictionary with attribute-style access'''



# -*- coding: utf-8 -*-
from __future__ import absolute_import

from .base import stuf
from .core import defaultstuf, fixedstuf, frozenstuf, orderedstuf

idefaultstuf = defaultstuf
ifixedstuf = fixedstuf
ifrozenstuf = frozenstuf
iorderedstuf = orderedstuf
istuf = stuf



# -*- coding: utf-8 -*-
## pylint: disable-msg=w0702
'''stuf utilities'''

from __future__ import absolute_import
import inspect
try:
    from thread import get_ident
except ImportError:
    from dummy_thread import get_ident
try:
    from collections import OrderedDict
except  ImportError:
    from ordereddict import OrderedDict
from functools import wraps, update_wrapper


def class_name(this):
    '''
    get class name

    @param this: object
    '''
    return getattr(this.__class__, '__name__')


def deleter(this, key):
    '''
    delete an attribute

    @param this: object
    @param key: key to lookup
    '''
    if inspect.isclass(this):
        delattr(this, key)
    else:
        object.__delattr__(this, key)


def getter(this, key, default=None):
    '''
    get an attribute

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    if inspect.isclass(this):
        return getattr(this, key, default)
    return this.__dict__.get(key, default)


def lru_wrapped(this, maxsize=100):
    '''
    least-recently-used cache decorator from Raymond Hettinger

    arguments to the cached function must be hashable.

    @param maxsize: maximum number of results in LRU cache (default: 100)
    '''
    # order: least recent to most recent
    cache = OrderedDict()

    @wraps(this)
    def wrapper(*args, **kw):
        key = args
        if kw:
            key += tuple(sorted(kw.items()))
        try:
            result = cache.pop(key)
        except KeyError:
            result = this(*args, **kw)
            # purge least recently used cache entry
            if len(cache) >= maxsize:
                cache.popitem(0)
        # record recent use of this key
        cache[key] = result
        return result
    return wrapper


def object_name(this):
    '''
    get object name

    @param this: object
    '''
    return getattr(this, '__name__')


def recursive_repr(this):
    '''
    Decorator to make a repr function return "..." for a recursive call

    @param this: object
    '''
    repr_running = set()

    def wrapper(self):
        key = id(self), get_ident()
        if key in repr_running:
            return '...'
        repr_running.add(key)
        try:
            result = this(self)
        finally:
            repr_running.discard(key)
        return result
    # Can't use functools.wraps() here because of bootstrap issues
    wrapper.__module__ = getattr(this, '__module__')
    wrapper.__doc__ = getattr(this, '__doc__')
    wrapper.__name__ = object_name(this)
    return wrapper


def setter(this, key, value):
    '''
    get an attribute

    @param this: object
    @param key: key to set
    @param value: value to set
    '''
    if inspect.isclass(this):
        setattr(this, key, value)
    else:
        this.__dict__[key] = value


class lazybase(object):

    def __init__(self, method):
        self.method = method
        try:
            self.__doc__ = method.__doc__
            self.__module__ = method.__module__
            self.__name__ = self.name = object_name(method)
        except:
            pass


class lazy(lazybase):

    '''Lazily assign attributes on an instance upon first use.'''

    def __get__(self, instance, owner):
        if instance is None:
            return self
        value = self.method(instance)
        setter(instance, self.name, value)
        return value


class lazy_class(lazybase):

    '''Lazily assign attributes on an class upon first use.'''

    def __get__(self, instance, owner):
        value = self.method(owner)
        setter(owner, self.name, value)
        return value
    

class both(lazy):

    '''
    decorator which allows definition of a Python descriptor with both
    instance-level and class-level behavior
    '''

    def __init__(self, method, expr=None):
        super(both, self).__init__(method)
        self.expr = expr or method
        update_wrapper(self, method)

    def __get__(self, instance, owner):
        if instance is None:
            return self.expr(owner)
        return super(both, self).__get__(instance, owner)

    def expression(self, expr):
        '''
        a modifying decorator that defines a general method
        '''
        self.expr = expr
        return self


class either(both):

    '''
    decorator which allows definition of a Python descriptor with both
    instance-level and class-level behavior
    '''

    def __get__(self, instance, owner):
        if instance is None:
            value = self.expr(owner)
            setter(owner, self.name, value)
            return value
        value = self.method(instance)
        setter(instance, self.name, value)
        return value



# -*- coding: utf-8 -*-
# pylint: disable-msg=w0221,w0212,w0201
'''core stuf'''

from __future__ import absolute_import
from collections import (
    MutableMapping, Mapping, Sequence, defaultdict, namedtuple,
)

from .util import OrderedDict
from .base import basestuf, wrapstuf


class defaultstuf(basestuf, defaultdict):

    '''
    dictionary with attribute-style access and a factory function to provide a
    default value for keys with no value
    '''

    _mapping = defaultdict

    def __init__(self, default, *args, **kw):
        '''
        @param default: function that can provide default values
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        defaultdict.__init__(self, default)
        basestuf.__init__(self, *args, **kw)

    @classmethod
    def _build(cls, default, iterable):
        kind = cls._mapping
        # add class to handle potential nested objects of the same class
        kw = kind(default)
        if isinstance(iterable, Mapping):
            kw.update(kind(default, iterable))
        elif isinstance(iterable, Sequence):
            # extract appropriate key-values from sequence
            for arg in iterable:
                try:
                    kw.update(arg)
                except (ValueError, TypeError):
                    pass
        return kw

    @classmethod
    def _new(cls, default, iterable):
        return cls(default, cls._build(default, iterable))

    def _populate(self, iterable):
        new = self._new
        for k, v in iterable.items():
            if isinstance(v, (Mapping, Sequence)):
                # see if stuf can be converted to nested stuf
                trial = new(self.default_factory, v)
                if len(trial) > 0:
                    self[k] = trial
                else:
                    self[k] = v
            else:
                self[k] = v

    def _prepare(self, *args, **kw):
        kw.update(self._build(self.default_factory, args))
        return kw

    def copy(self):
        return self._build(self.default_factory, dict(i for i in self))

    def update(self, *args, **kw):
        self._populate(self._prepare(self.default_factory, *args, **kw))


class frozenstuf(Mapping):

    '''immutable dictionary with attribute-style access'''

    __slots__ = ['_items']

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        self._items = self._populate(self._build(dict(*args, **kw)))

    def __reduce__(self):
        return self.__class__, (self._items._asdict().copy(),)

    def __getitem__(self, key):
        try:
            return getattr(self._items, key)
        except AttributeError:
            raise KeyError('key not found')

    def __getattr__(self, key):
        try:
            return object.__getattribute__(self, key)
        except AttributeError:
            return getattr(self._items, key)

    def __iter__(self):
        return self._items._asdict().iterkeys()

    def __len__(self):
        return len(self._items)

    @classmethod
    def _build(cls, iterable):
        kw = dict()
        if isinstance(iterable, Mapping):
            kw.update(dict(i for i in iterable.items()))
        elif isinstance(iterable, Sequence):
            # extract appropriate key-values from sequence
            for arg in iterable:
                try:
                    kw.update(arg)
                except (ValueError, TypeError):
                    pass
        return kw

    def _freeze(self, mapping):
        frozen = namedtuple('frozenstuf', mapping.keys())
        return frozen(**mapping)

    def _populate(self, iterable):
        new = self._freeze
        for k, v in iterable.iteritems():
            if isinstance(v, basestring):
                iterable[k] = v
            elif isinstance(v, (Sequence, Mapping)):
                # see if stuf can be converted to nested stuf
                trial = frozenstuf(v)
                if len(trial) > 0:
                    iterable[k] = trial
                else:
                    iterable[k] = v
            else:
                iterable[k] = v
        return new(iterable)

    def copy(self):
        return self._build(dict(i for i in self.iteritems()))


class orderedstuf(wrapstuf, MutableMapping):

    '''dictionary with dot attributes that remembers insertion order'''

    _mapping = OrderedDict

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        self._wrapped = OrderedDict()
        super(orderedstuf, self).__init__(*args, **kw)

    def __setitem__(self, key, value):
        self._wrapped[key] = value

    def __delitem__(self, key):
        del self._wrapped[key]

    def __reversed__(self):
        return self._wrapped.__reversed__()

    def __reduce__(self):
        return self._wrapped.__reduce__()

    def copy(self):
        return self._build(OrderedDict(self))


class fixedstuf(wrapstuf, MutableMapping):

    '''
    dictionary with attribute-style access with mutability restricted to
    initial keys
    '''

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        self._wrapped = dict()
        super(fixedstuf, self).__init__(*args, **kw)

    def __setitem__(self, k, v):
        # only access initial keys
        if k in self.allowed:
            self._wrapped[k] = v
        else:
            raise KeyError('%s is not an allowed key' % k)

    def __setattr__(self, k, v):
        # allow normal object creation for protected keywords
        if k == '_classkeys' or k in self._classkeys:
            object.__setattr__(self, k, v)
        else:
            try:
                self[k] = v
            except KeyError:
                raise AttributeError(k)

    def __delitem__(self, key):
        self._wrapped[key] = None

    def __delattr__(self, k):
        self.__delitem__(k)

    def __reduce__(self):
        return self.__class__, (self._wrapped.copy(),)

    def _populate(self, iterable):
        self.allowed = frozenset(iterable.iterkeys())
        super(fixedstuf, self)._populate(iterable)

    def copy(self):
        return fixedstuf(self._wrapped)

    def clear(self):
        self._wrapped.clear()

    def popitem(self):
        raise AttributeError()

    def pop(self, key, default=None):
        raise AttributeError()
